# encoding: utf-8
# pylint: disable=pointless-string-statement

from __future__ import unicode_literals
import sys


class FullTraceback(object):
    def __init__(self, tb_frame, tb_lineno, tb_next):
        self.tb_frame = tb_frame
        self.tb_lineno = tb_lineno
        self.tb_next = tb_next


def current_stack(skip=0):
    p_frame = None
    try:
        1/0
    except ZeroDivisionError:
        p_frame = sys.exc_info()[2].tb_frame

    for _ in xrange(skip + 2):
        p_frame = p_frame.f_back
    rg_frames = []
    while p_frame is not None:
        rg_frames.append((p_frame, p_frame.f_lineno))
        p_frame = p_frame.f_back
    return rg_frames


def extend_traceback(tb, stack, n_max=6):
    """
    This function "extends" a traceback by filling in
    actual stack frames into the traceback.  This is useful
    for debugging and for using the full traceback with
    Sentry
    """
    p_head = tb
    for x, (tb_frame, tb_lineno) in enumerate(stack):
        if x >= n_max:
            break
        p_head = FullTraceback(tb_frame, tb_lineno, p_head)
    return p_head


def full_exc_info():
    """
    This function is designed to mimic sys.exc_info.
    However, unlike sys.exc_info, the `full_exc_info`
    function returns the full traceback.
    """
    t, v, p_traceback = sys.exc_info()
    p_full = extend_traceback(p_traceback, current_stack(1))
    return t, v, p_full


# This file uses the following encoding:  utf-8


import logging
from django.core.management.base import BaseCommand
from commandlogginglevels import VerbosityLevels as V
import socket
from djenga.mixins.loggingmixin import LoggingMixin
from djenga.email.helpers import send_html_email


class EmailCommand(BaseCommand, LoggingMixin):
    help = u'%s%s%s%s%s' % (
        u'djenga.management.command.EmailCommand is a ',
        u'base class that provides useful functionality for ',
        u'django management commands such as logging and ',
        u'e-mail.  Subclasses should override the _handle method ',
        u'to implement actual commands.',
    )

    def _handle(self, *args, **options):
        pass

    def handle(self, *args, **options):
        self.verbosity = options.get(u'verbosity', 1)
        self.logging_level = V.to_logging_level(self.verbosity)
        return_value = self._handle(*args, **options)
        self._send_email()
        return return_value

    def _send_email(self):
        pass

    def _send_html_email(self, html_template, context, text_template=None):
        context[u'hostname'] = socket.gethostname()
        context[u'summary_messages'] = self.log_map.get(logging.INFO, [])
        context[u'info_messages'] = self.log_map.get(logging.DEBUG, [])
        context[u'warning_messages'] = self.log_map.get(logging.WARNING, [])
        context[u'error_messages'] = self.log_map.get(
            logging.ERROR, []
        ) + self.log_map.get(
            logging.CRITICAL, []
        )
        send_html_email(
            subject=u'%s' % (self.__class__.__module__,),
            template=html_template,
            context=context,
            text_template=text_template,
            inline_css=True,
        )



import inspect
import yaml
from time import time
from djenga.utils import flush_print


class IntegrationTestException(Exception):
    pass


class IntegrationTest(object):
    """
    Subclass this class to use for running integration
    tests with the IntegrationTestCommand.  In order
    to use this class, a subclass should define one or
    more instance member functions that begin with the
    word `test` and takes no arguments other than self.
    These functions will be called in alphabetical order
    by the super classes `run_test` function.  Example:

    from djenga.test import IntegrationTest

    class MyIntegrationTest(IntegrationTest):
        def test_addition(self):
            x = 1
            self.assert_equal(x, 1)
            x += 4
            self.assert_equal(x, 5)

    An integration test can also access YAML or JSON test
    data from a file specified by the class variable
    `test_data_file`.  This test data will is made
    available to the sub class through `self.test_data`.
    Example:
        In a file called path/to/my_test_data.yml

        ---
        step_1: 1
        step_2: 5

        In the integration test file:

        class MyIntegrationTest(IntegrationTest):
            test_data_file = 'path/to/my_test_data.yml'

            def test_addition(self):
                x = 1
                self.assert_equal(
                    x, self.test_data['step_1'],
                    'Oh no! failed step 1')
                x += 4
                self.assert_equal(
                    x, self.test_data['step_2'],
                    'Oh no! failed step 2')
    """
    test_data_file = ''

    def __init__(self):
        self.test_data = None
        self.load_test_data()

    def load_test_data(self):
        if not self.test_data_file:
            return
        with open(self.test_data_file, 'r') as f:
            self.test_data = yaml.load(f)

    def assert_equal(self, left, right, message=''):
        if left != right:
            raise IntegrationTestException(message)

    def run_test(self):
        test_methods = [
            value
            for key, value in inspect.getmembers(self)
            if key.startswith('test') and callable(value)
        ]
        test_methods.sort(key=lambda fn: fn.__name__)
        all_passed = True
        for x in test_methods:
            flush_print('%s.%s........', self.__class__.__name__, x.__name__, end='')
            tm_start = time()
            status = 'passed'
            units = 's'
            try:
                x()
            except IntegrationTestException as ex:
                status = 'failed'
                if ex.message:
                    status += ' (%s)' % ex.message
                all_passed = False
            finally:
                tm_total = time() - tm_start
            if tm_total < 1:
                tm_total *= 1000
                units = 'ms'
            flush_print('%s %.1f%s', status, tm_total, units)
        return all_passed



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-04-09 20:27
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='CommandOutput',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('output', models.TextField(null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name='ManagementCommand',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, help_text=b'The full path of the command', max_length=128)),
                ('status', models.CharField(help_text=b'The current status of the command', max_length=32)),
                ('last_run', models.DateTimeField(help_text=b'The last time this command was run', null=True)),
                ('last_success', models.DateTimeField(help_text=b'The last time this command was run successfully', null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.AddField(
            model_name='commandoutput',
            name='command',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djenga.ManagementCommand'),
        ),
    ]



"""
# Included with pynliner since it isn't on PyPI #

soupselect.py

CSS selector support for BeautifulSoup.

soup = BeautifulSoup('<html>...')
select(soup, 'div')
    - returns a list of div elements

select(soup, 'div#main ul a')
    - returns a list of links inside a ul inside div#main

patched to support multiple class selectors here http://code.google.com/p/soupselect/issues/detail?id=4#c0
"""
import re
import bs4

attribute_regex = re.compile('\[(?P<attribute>\w+)(?P<operator>[=~\|\^\$\*]?)=?["\']?(?P<value>[^\]"]*)["\']?\]')
pseudo_classes_regexes = (
    re.compile(':(first-child)'),
    re.compile(':(last-child)')
)

def get_attribute_checker(operator, attribute, value=''):
    """
    Takes an operator, attribute and optional value; returns a function that
    will return True for elements that match that combination.
    """
    return {
        '=': lambda el: el.get(attribute) == value,
        # attribute includes value as one of a set of space separated tokens
        '~': lambda el: value in el.get(attribute, '').split(),
        # attribute starts with value
        '^': lambda el: el.get(attribute, '').startswith(value),
        # attribute ends with value
        '$': lambda el: el.get(attribute, '').endswith(value),
        # attribute contains value
        '*': lambda el: value in el.get(attribute, ''),
        # attribute is either exactly value or starts with value-
        '|': lambda el: el.get(attribute, '') == value \
            or el.get(attribute, '').startswith('%s-' % value),
    }.get(operator, lambda el: el.has_key(attribute))

def is_white_space(el):
    if isinstance(el, BeautifulSoup.NavigableString) and str(el).strip() == '':
        return True
    if isinstance(el, BeautifulSoup.Comment):
        return True
    return False

def is_last_content_node(el):
    result = False
    if el is None:
        result = True
    elif is_white_space(el):
        result = is_last_content_node(el.nextSibling)
    return result

def is_first_content_node(el):
    result = False
    if el is None:
        result = True
    if is_white_space(el):
        result = is_first_content_node(el.previousSibling)
    return result
        
def get_pseudo_class_checker(psuedo_class):
    """
    Takes a psuedo_class, like "first-child" or "last-child"
    and returns a function that will check if the element satisfies
    that psuedo class
    """
    return {
        'first-child': lambda el: is_first_content_node(getattr(el, 'previousSibling', None)),
        'last-child': lambda el: is_last_content_node(getattr(el, 'nextSibling', None))
    }.get(psuedo_class)

def get_checker(functions):
    def checker(el):
        for func in functions:
            if not func(el):
                return False
        return el
    return checker


def select(soup, selector):
    """
    soup should be a BeautifulSoup instance; selector is a CSS selector 
    specifying the elements you want to retrieve.
    """
    handle_token = True
    current_context = [(soup, [])]
    operator = None
    while selector:
        if handle_token:
            # Get the rightmost token
            handle_token = False
            match = re.search('([_0-9a-zA-Z-#.:*"\'\[\\]=]+)$', selector)
            if not match:
                raise Exception("No match was found. We're done or something is broken")
            token = match.groups(1)[0]

            # remove this token from the selector
            selector = selector.rsplit(token, 1)[0].rstrip()
            
            checker_functions = []
            #
            # Get attribute selectors from token
            #
            matches = attribute_regex.findall(token)
            for match in matches:
                checker_functions.append(get_attribute_checker(match[1], match[0], match[2]))

            #
            # Get pseudo classes from token
            #
            for pseudo_class_regex in pseudo_classes_regexes:
                match = pseudo_class_regex.search(token)
                if match:
                    checker_functions.append(get_pseudo_class_checker(match.groups(1)[0]))

            checker = get_checker(checker_functions)
            #
            # Get tag
            #
            tag = re.findall('^([a-zA-Z0-9]+)', token)
            if len(tag) == 0:
                tag = True
            elif len(tag) == 1:
                tag = tag[0]
            else:
                raise Exception("Multiple tags found (invalid CSS)")

            #
            # Get ID
            #
            ids = re.findall('#([a-zA-Z0-9_-]+)', token)
            if len(ids) > 1:
                raise Exception("Only single # OK")
            #
            # Get classes
            #
            classes = re.findall('\.([a-zA-Z0-9_-]+)', token)

            #
            # Search contexts for matches
            #
            found = []
            find_dict = {}
            if ids:
                find_dict['id'] = ids
            if classes:
                find_dict['class'] = lambda attr: attr and set(classes).issubset(attr.split())
            if operator is None:
                # This is the first token: simply find all matches
                for context in current_context:
                    context_matches = [el for el in context[0].findAll(tag, find_dict) if checker(el)]
                    for context_match in context_matches:
                        found.append(
                            (context_match, [context_match]),
                        )
            elif operator == ' ':
                # for each context in current_context, ensure there
                # exists an element somewhere above that element that
                # matches the provided token
                # ("descendant" selector)
                for context in current_context:
                    context_matches = []
                    for el in context[1]:
                        if checker(el.findParent(tag, find_dict)):
                            context_matches.append(el)
                    if context_matches:
                        found.append(
                            (context[0], context_matches),
                        )
            elif operator == '>':
                # for each context in current_context,
                # check if the parent satisfies the provided
                # arguments.
                for context in current_context:
                    context_matches = []
                    for el in context[1]:
                        if checker(el.findParent(tag, find_dict)) == el.parent:
                            context_matches.append(el.parent)
                    if context_matches:
                        found.append(
                            (context[0], context_matches),
                        )
            elif operator == '~':
                # for each context in current_context
                # check 
                raise NotImplementedError("~ operator is not implemented. Sad face :(")
            elif operator == '+':
                # for each context in current_context
                # check if the preceding sibling satisfies the
                # provided arguments
                for context in current_context:
                    context_matches = []
                    for el in context[1]:
                        if checker(el.findPreviousSibling(tag, find_dict)) == el.previousSibling:
                            context_matches.append(el.previousSibling)
                    if context_matches:
                        found.append(
                            (context[0], context_matches)
                        )
            current_context = found
        else:
            # Get the next operator (whitespace, >, ~, +)
            handle_token = True
            operator = None
            match = re.search('([>~+]+)$', selector)
            if match:
                operator = match.groups(1)[0]
            else:
                operator = ' '
            selector = selector.rsplit(operator, 1)[0].rstrip()
    return [entry[0] for entry in current_context]

def monkeypatch(BeautifulSoupClass=None):
    """
    If you don't explicitly state the class to patch, defaults to the most 
    common import location for BeautifulSoup.
    """
    if not BeautifulSoupClass:
        from bs4 import BeautifulSoup as BeautifulSoupClass
    BeautifulSoupClass.findSelect = select

def unmonkeypatch(BeautifulSoupClass=None):
    if not BeautifulSoupClass:
        from bs4 import BeautifulSoup as BeautifulSoupClass
    delattr(BeautifulSoupClass, 'findSelect')



from unicode_csv_reader import read_unicode_csv
from unicode_csv_reader import UnicodeCsvReader
from unicode_csv_writer import UnicodeCsvWriter


__all__ = [
    'read_unicode_csv',
    'UnicodeCsvReader',
    'UnicodeCsvWriter',
]



import codecs
from djenga.csv import UnicodeCsvWriter


class CsvMixin(object):
    """
    Enhances a django management command by allowing it to
    write to a CSV file through an
    instance function called `writerow`.  The output file
    may be specified at the command line by using the
    `--output` or `-o` option.  If no output file is specified,
    the command will output to stdout.

    n.b. This mixin requires that any sub-classes making use of
    it to call super(ClassName, self).add_arguments(parser)
    at the beginning of any override of the `add_arguments`
    method.
    """
    output_file = None

    def __init__(self):
        super(CsvMixin, self).__init__()
        self.csv_writer = None
        """@type: UnicodeCsvWriter"""
        self.file_handle = None
        """@type: stream"""

    def add_arguments(self, parser):
        """
        Adds the output file argument to the list of
        arguments accepted by the command
        :type parser: argparse.ArgumentParser
        """
        super(CsvMixin, self).add_arguments(parser)
        parser.add_argument(
            '-o', '--output',
            dest='output_file',
            help='The /path/to/output-file.csv of the output CSV.'
                 '  If no output file is specified, the command'
                 ' will output the CSV contents to stdout.')

    def get_csv_writer(self):
        """
        Returns an initialized CSV writer
        :rtype: UnicodeCsvWriter
        """
        if not self.csv_writer:
            if self.output_file:
                self.file_handle = codecs.open(self.output_file, 'w', encoding='utf-8')
            else:
                self.file_handle = self.stdout
            self.csv_writer = UnicodeCsvWriter(self.file_handle)
        return self.csv_writer

    def writerow(self, data):
        """
        Writes a row of data to the CSV
        :param data: a row of data to be written
        """
        self.get_csv_writer().writerow(data)

    def close_file_handle(self):
        """
        Closes any initialized file handles
        """
        if self.file_handle and self.output_file:
            self.file_handle.close()

    def execute(self, *args, **options):
        """
        Override the execute function so that we
        can get a handle on the output_file argument
        and close the file handle gracefully after command
        exit.
        """
        self.output_file = options.get('output_file')
        try:
            super(CsvMixin, self).execute(*args, **options)
        finally:
            self.close_file_handle()




from djenga.management.commands import StatusCommand


class Command(StatusCommand):
    def handle(self, *args, **options):
        self.debug('Hello Sesame')
        self.info('Hello Stella')
        self.warning('Hello Penny')
        return ''




# encoding: utf-8

import pytz
from django.db import models


class TimestampField(models.DateTimeField):
    def from_db_value(self, value, expression, connection, context):
        if value is None or getattr(value, 'tzinfo', None) is not None:
            return value
        value = pytz.utc.localize(value)
        return value





import os
import json
import yaml


class ConfigurationMixin(object):
    """
    Enhances a django management command by allowing it to
    read from a configuration file at the beginning of its
    run and exposing the loaded configuration in an
    instance member called `config`.  The default
    configuration file may be specified in a class member
    called `config_file`.  This default configuration file
    may be overridden at the command line by using the
    `--config` or `-c` option.

    n.b. This mixin requires that any sub-classes making use of
    it to call super(ClassName, self).add_arguments(parser)
    at the beginning of any override of the `add_arguments`
    method.
    """
    config_file = None

    def __init__(self):
        super(ConfigurationMixin, self).__init__()
        self.config = {}

    def add_arguments(self, parser):
        """
        :type parser: argparse.ArgumentParser
        """
        super(ConfigurationMixin, self).add_arguments(parser)
        parser.add_argument(
            '-c', '--config',
            dest='config_file',
            default=self.config_file,
            help='the /path/to the configuration file')

    def load_config(self):
        if not self.config_file:
            return
        _, ext = os.path.splitext(self.config_file)
        loader = json
        if ext in ('.yml', '.yaml'):
            loader = yaml
        with open(self.config_file, 'r') as f:
            self.config = loader.load(f)

    def execute(self, *args, **options):
        self.config_file = options.get('config_file')
        self.load_config()
        super(ConfigurationMixin, self).execute(*args, **options)



# encoding: utf-8


from __future__ import unicode_literals
import logging
import codecs
import sys
from django.conf import settings
from django.core.management.base import OutputWrapper


class LoggingMixin(object):
    verbosity = 3 if settings.DEBUG else 1
    """@type: int"""
    indent = 0
    """@type: int"""
    logging_level = logging.DEBUG if settings.DEBUG else 1
    log_map = dict()
    logging_initialized = False
    print_level = True

    def set_verbosity(self, verbosity):
        LEVELS = {
            0: logging.CRITICAL,
            1: logging.ERROR,
            2: logging.WARNING,
            3: logging.DEBUG,
        }
        self.verbosity = verbosity
        self.logging_level = LEVELS[verbosity]

    def initialize_logging(self):
        if not self.logging_initialized:
            self.stdout = OutputWrapper(
                getattr(self, 'stdout', sys.stdout),
                ending='')
            self.stdout = codecs.getwriter('utf8')(self.stdout)
            self.logging_initialized = True

    def color_format(self, level, message):
        level_colors = {
            # Level and a pair of colors: first for the label, the rest for the text;
            #   the bolder color label can make them easier to spot in the console log.
            logging.DEBUG:        ( 33,  39),
            # logging.TRACE:        (147, 153),
            logging.INFO:         ( 43,  49),
            logging.WARNING:      (214, 226),
            logging.ERROR:        (196, 197),
            logging.CRITICAL:     (196, 197),
        }.get(level, (33, 39))
        color   = "\033[38;5;{:d}m"         # 256-color to give wider spectrum than just ANSI
        reset   = "\033[0m"

        # Pass any simple messages from internal things, like Django's runserver, without special formatting.
        mp_levels = {
            logging.INFO: u'INF',
            logging.WARNING: u'WRN',
            logging.ERROR: u'ERR',
            logging.DEBUG: u'DBG',
            logging.CRITICAL: u'CRT'
        }
        st_level = mp_levels[level]
        level_prefix = '%s[%s] ' % (color.format(level_colors[0]), st_level)
        return u'{level_prefix}{color_normal}{message}{reset}'.format(
            level_prefix    = level_prefix if self.print_level else '',
            message         = message,
            color_normal    = color.format(level_colors[1]),
            reset           = reset
        )

    def llog(self, logging_level, format_string, *args):
        """
        @param logging_level:
            50 = summary/critical
            40 = error
            30 = warning
            20 = info
            10 = debug
        @return:
        """
        LEVELS = {
            logging.CRITICAL,
            logging.ERROR,
            logging.WARNING,
            logging.INFO,
            logging.DEBUG
        }
        if logging_level not in LEVELS:
            logging_level = logging.DEBUG
        message = format_string % args
        if logging_level >= self.logging_level:
            if hasattr(self, 'stdout'):
                self.initialize_logging()
                self.stdout.write(u' ' * self.indent)
                if self.stdout.isatty():
                    self.stdout.write(self.color_format(logging_level, message))
                else:
                    self.stdout.write(message)
                self.stdout.write('\n')
                self.log_map.setdefault(logging_level, []).append(message)

    def log(self, format_string, *args):
        message = format_string % args
        if hasattr(self, 'stdout'):
            self.initialize_logging()
            self.stdout.write(u' ' * self.indent)
            self.stdout.write(message)
            self.stdout.write('\n')

    def critical(self, format_string, *args):
        self.llog(logging.CRITICAL, format_string, *args)

    def debug(self, format_string, *args):
        self.llog(logging.DEBUG, format_string, *args)

    def info(self, format_string, *args):
        self.llog(logging.INFO, format_string, *args)

    def warning(self, format_string, *args):
        self.llog(logging.WARNING, format_string, *args)

    def error(self, format_string, *args):
        self.llog(logging.ERROR, format_string, *args)

    def exception(self, format_string, *args):
        p_type, p_exception, p_traceback = sys.exc_info()
        self.llog(logging.ERROR, format_string, *args)
        self.llog(logging.ERROR, u'Exception message: %s', p_exception)
        self.llog(logging.ERROR, u'Exception type   : %s', p_type)
        self.llog(logging.ERROR, u'Traceback\n%s', p_traceback.format_exc())







#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Pynliner : Convert CSS to inline styles

Python CSS-to-inline-styles conversion tool for HTML using BeautifulSoup and
cssutils

Copyright (c) 2011-2013 Tanner Netterville

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

The generated output of this software shall not be used in a mass marketing
service.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR 
THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

__version__ = "0.5.0"

import re
import urlparse
import urllib2
import cssutils
from bs4 import BeautifulSoup
from soupselect import select


class Pynliner(object):
    """Pynliner class"""

    soup = False
    style_string = False
    stylesheet = False
    output = False

    def __init__(self, log=None, allow_conditional_comments=False):
        self.log = log
        cssutils.log.enabled = False if log is None else True
        self.extra_style_strings = []
        self.allow_conditional_comments = allow_conditional_comments
        self.root_url = None
        self.relative_url = None

    def from_url(self, url):
        """Gets remote HTML page for conversion

        Downloads HTML page from `url` as a string and passes it to the
        `from_string` method. Also sets `self.root_url` and `self.relative_url`
        for use in importing <link> elements.

        Returns self.

        >>> p = Pynliner()
        >>> p.from_url('http://somewebsite.com/file.html')
        <Pynliner object at 0x26ac70>
        """
        self.url = url
        self.relative_url = '/'.join(url.split('/')[:-1]) + '/'
        self.root_url = '/'.join(url.split('/')[:3])
        self.source_string = self._get_url(self.url)
        return self

    def from_string(self, string):
        """Generates a Pynliner object from the given HTML string.

        Returns self.

        >>> p = Pynliner()
        >>> p.from_string('<style>h1 {color:#ffcc00;}</style><h1>Hi</h1>')
        <Pynliner object at 0x26ac70>
        """
        self.source_string = string
        return self

    def with_cssString(self, css_string):
        """Adds external CSS to the Pynliner object. Can be "chained".

        Returns self.

        >>> html = "<h1>Hello World!</h1>"
        >>> css = "h1 { color:#ffcc00; }"
        >>> p = Pynliner()
        >>> p.from_string(html).with_cssString(css)
        <pynliner.Pynliner object at 0x2ca810>
        """
        self.extra_style_strings.append(css_string)
        return self

    def run(self):
        """Applies each step of the process if they have not already been
        performed.

        Returns Unicode output with applied styles.

        >>> html = "<style>h1 { color:#ffcc00; }</style><h1>Hello World!</h1>"
        >>> Pynliner().from_string(html).run()
        u'<h1 style="color: #fc0">Hello World!</h1>'
        """
        if not self.soup:
            self._get_soup()
        if not self.stylesheet:
            self._get_styles()
        self._apply_styles()
        self._get_output()
        self._clean_output()
        return self.output

    def _get_url(self, url):
        """Returns the response content from the given url
        """
        return urllib2.urlopen(url).read()

    def _get_soup(self):
        """Convert source string to BeautifulSoup object. Sets it to self.soup.

        If using mod_wgsi, use html5 parsing to prevent BeautifulSoup
        incompatibility.
        """
        # Check if mod_wsgi is running
        # - see http://code.google.com/p/modwsgi/wiki/TipsAndTricks
        try:
            from mod_wsgi import version
            self.soup = BeautifulSoup(self.source_string, "html5lib")
        except:
            self.soup = BeautifulSoup(self.source_string)

    def _get_styles(self):
        """Gets all CSS content from and removes all <link rel="stylesheet"> and
        <style> tags concatenating into one CSS string which is then parsed with
        cssutils and the resulting CSSStyleSheet object set to
        `self.stylesheet`.
        """
        self._get_external_styles()
        self._get_internal_styles()
        for style_string in self.extra_style_strings:
            self.style_string += style_string
        cssparser = cssutils.CSSParser(log=self.log)
        self.stylesheet = cssparser.parseString(self.style_string)

    def _get_external_styles(self):
        """Gets <link> element styles
        """
        if not self.style_string:
            self.style_string = u''
        else:
            self.style_string += u'\n'

        link_tags = self.soup.findAll('link', {'rel': 'stylesheet'})
        for tag in link_tags:
            url = tag['href']

            # Convert the relative URL to an absolute URL ready to pass to urllib
            base_url = self.relative_url or self.root_url
            url = urlparse.urljoin(base_url, url)

            self.style_string += self._get_url(url)
            tag.extract()

    def _get_internal_styles(self):
        """Gets <style> element styles
        """
        if not self.style_string:
            self.style_string = u''
        else:
            self.style_string += u'\n'

        style_tags = self.soup.findAll('style')
        for tag in style_tags:
            self.style_string += u'\n'.join(tag.contents) + u'\n'
            tag.extract()

    def _get_specificity_from_list(self, lst):
        """
        Takes an array of ints and returns an integer formed
        by adding all ints multiplied by the power of 10 of the current index

        (1, 0, 0, 1) => (1 * 10**3) + (0 * 10**2) + (0 * 10**1) + (1 * 10**0) => 1001
        """
        return int(''.join(map(str, lst)))

    def _get_rule_specificity(self, rule):
        """
        For a given CSSRule get its selector specificity in base 10
        """
        return sum(map(self._get_specificity_from_list, (s.specificity for s in rule.selectorList)))

    def _apply_styles(self):
        """Steps through CSS rules and applies each to all the proper elements
        as @style attributes prepending any current @style attributes.
        """
        rules = self.stylesheet.cssRules.rulesOfType(1)
        elem_props = list()

        # build up a property list for every styled element
        for rule in rules:
            # select elements for every selector
            selectors = rule.selectorText.split(',')
            elements = []
            for selector in selectors:
                elements_found = self.soup.select(selector)
                elements += elements_found
            # build prop_list for each selected element
            for elem in elements:
                elem_props.append((
                    elem,
                    self._get_rule_specificity(rule),
                    rule.style.getProperties(),
                ))

        # build up another property list using selector specificity
        elem_styles = list()
        for elem, _, props in elem_props:
            new_style = cssutils.css.CSSStyleDeclaration()
            for prop in props:
                new_style.setProperty(prop.name, prop.value)
            elem_styles.append((elem, new_style))

        for elem, style_declaration in elem_styles:
            if elem.has_attr('style'):
                elem['style'] = u'%s; %s' % (style_declaration.cssText.replace('\n', ' '), elem['style'])
            else:
                elem['style'] = style_declaration.cssText.replace('\n', ' ')

    def _get_output(self):
        """Generate Unicode string of `self.soup` and set it to `self.output`

        Returns self.output
        """
        self.output = unicode(self.soup)
        return self.output
    
    def _clean_output(self):
        """Clean up after BeautifulSoup's output.
        """
        if self.allow_conditional_comments:
            matches = re.finditer('(<!--\[if .+\].+?&lt;!\[endif\]-->)', self.output)
            for match in matches:
                comment = match.group()
                comment = comment.replace('&gt;', '>')
                comment = comment.replace('&lt;', '<')
                self.output = (self.output[:match.start()] + comment +
                               self.output[match.end():])


def fromURL(url, log=None):
    """Shortcut Pynliner constructor. Equivalent to:

    >>> Pynliner().from_url(someURL).run()

    Returns processed HTML string.
    """
    return Pynliner(log).from_url(url).run()


def from_string(string, log=None):
    """Shortcut Pynliner constructor. Equivalent to:

    >>> Pynliner().from_string(someString).run()

    Returns processed HTML string.
    """
    return Pynliner(log).from_string(string).run()



import pynliner


class EngineBase(object):
    def __init__(self, html, css):
        self.html = html
        self.css = css

    def render(self):
        raise NotImplementedError()


class InlineCssEngine(EngineBase):
    def render(self):
        inliner = pynliner.Pynliner().from_string(self.html)
        inliner = inliner.with_cssString(self.css)
        return inliner.run()


class NullEngine(EngineBase):
    def render(self):
        return self.html




# encoding: utf-8

from __future__ import unicode_literals
from uuid import UUID
from django.db.backends.mysql.base import django_conversions
from django.db import models


def prep_uuid(
        o,
        *args # pylint: disable=unused-argument
):
    return '0x%s' % o.hex


django_conversions.update({
    UUID: prep_uuid
})


class UuidField(models.fields.Field):
    """
    A Uuid field for MySQL (and MySQL only!) that converts to
    a `binary(16)` on the DB backend.  Unlike many of the implementations
    of binary fields out there (I'm looking at you django 1.6), this
    field does allow the user to do a lookup based on the UUID.

    The related value type of a Uuid field is the `uuid.UUID` class
    from python.
    """
    __metaclass__ = models.SubfieldBase

    def to_python(self, value):
        if isinstance(value, UUID):
            return value
        if value is None:
            return None
        return UUID(bytes=value)

    def get_prep_value(self, value):
        if value is None:
            return 'null'
        if isinstance(value, UUID):
            return value
        try:
            p = UUID(value)
            return p
        except ValueError:
            raise TypeError(
                'A %s cannot be used in a query involving a UUID' % (
                    value.__class__.__name__,
                )
            )

    def get_prep_lookup(self, lookup_type, value):
        if lookup_type not in { 'exact', 'iexact', 'isnull', 'in'}:
            raise TypeError('Binary Fields do not support %s' % lookup_type)
        return self.get_prep_value(value)

    def db_type(self, connection):
        if connection.settings_dict['ENGINE'] in { 'django.db.backends.mysql', 'django_mysqlpool.backends.mysqlpool' }:
            return 'binary(16)'
        else:
            raise Exception('MySql is the only defined engine for UuidField.')






class LookupEnum(object):
    @classmethod
    def values(cls):
        if hasattr(cls, 'mp_values'):
            return cls.mp_values
        cls.mp_values = {
            key: value for key, value
            in vars(cls).items() if not key.startswith('__')
        }
        return cls.mp_values

    @classmethod
    def value(cls, n_id):
        return cls.values().get(n_id)


class ReversibleEnum(object):
    @classmethod
    def codes(cls):
        if hasattr(cls, 'mp_codes'):
            return cls.mp_codes
        cls.mp_codes = {
            value: key for key, value
            in vars(cls).items() if not key.startswith('__')
        }
        return cls.mp_codes

    @classmethod
    def code(cls, n_id):
        return cls.codes().get(n_id)



# encoding: utf-8
from djenga.loggers.util import ColorFormatter
from djenga.loggers.util import BriefColorFormatter
from djenga.loggers.util import IgnoreCaching
from djenga.loggers.util import IgnoreDjangoInternals
from djenga.loggers.util import IgnorePisaPdf
from djenga.loggers.util import IgnoreRequests
from djenga.loggers.util import WorldWritableRotatingFileHandler



# encoding: utf-8


from __future__ import unicode_literals
from decimal import Decimal
from decimal import ROUND_FLOOR
from decimal import ROUND_DOWN
from decimal import ROUND_UP


__all__ = [
    u'currency_round_down',
    u'currency_round_up',
]


def currency_round_down(amount):
    """
    Useful helper function that takes a numerical amount, converts
    it to a decimal.Decimal object and rounds it down to the nearest
    cent.
    """
    amount = amount or 0.00
    amount = Decimal(amount)
    amount = amount.quantize(Decimal(u'.0001'), rounding=ROUND_FLOOR)
    amount = amount.quantize(Decimal(u'.01'), rounding=ROUND_DOWN)
    return amount


def currency_round_up(amount):
    """
    Useful helper function that takes a numerical amount, converts
    it to a decimal.Decimal object and rounds it up to the nearest
    cent.
    """
    amount = amount or 0.00
    amount = Decimal(amount)
    amount = amount.quantize(Decimal(u'.0001'), rounding=ROUND_FLOOR)
    amount = amount.quantize(Decimal(u'.01'), rounding=ROUND_UP)
    return amount




# encoding: utf-8


from __future__ import unicode_literals
import logging
import codecs
import sys
from datetime import datetime
from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import connections
from django.utils import timezone
from djenga.models import ManagementCommand
from djenga.models import CommandOutput


class StatusCommand(BaseCommand):
    def set_verbosity(self, verbosity):
        LEVELS = {
            0: logging.CRITICAL,
            1: logging.ERROR,
            2: logging.WARNING,
            3: logging.DEBUG,
        }
        self.verbosity = verbosity
        self.logging_level = LEVELS[verbosity]

    def create_parser(self, prog_name, subcommand):
        parser = super(StatusCommand, self).create_parser(prog_name, subcommand)
        parser.add_argument(
            '--last',
            action='store_true',
            dest='b_show_last',
            default=False
        )
        return parser

    def start_run(self):
        ManagementCommand.objects.update_or_create(
            name=self.command_name,
            defaults={
                'last_run': datetime.now(timezone.utc),
                'status': 'running',
            }
        )

    @property
    def command_name(self):
        name = self.__module__
        return name.replace('.management.commands', '')

    def end_run(self, success=True):
        if self.current_line:
            self.plain_log('\n')
        connection = connections['default']
        if connection.connection and not connection.is_usable():
            connection.close()
        q = ManagementCommand.objects.get(
            name=self.command_name,
        )
        q.status = 'success' if success else 'error'
        if success:
            q.last_success = datetime.now(timezone.utc)
        q.save()
        CommandOutput.objects.create(
            command=q,
            output='\n'.join(self.output)
        )

    def __init__(self):
        super(StatusCommand, self).__init__()
        self.verbosity = 3 if settings.DEBUG else 1
        """@type: int"""
        self.indent = 0
        """@type: int"""
        self.logging_level = logging.DEBUG if settings.DEBUG else 1
        self.output = []
        self.print_level = True
        self.stdout.ending = ''
        self.stdout = codecs.getwriter('utf8')(self.stdout)
        self.current_line = ''

    def color_format(self, level, message):
        level_colors = {
            # Level and a pair of colors: first for the label, the rest for the text;
            #   the bolder color label can make them easier to spot in the console log.
            logging.DEBUG:        ( 33,  39),
            # logging.TRACE:        (147, 153),
            logging.INFO:         ( 43,  49),
            logging.WARNING:      (214, 226),
            logging.ERROR:        (196, 197),
            logging.CRITICAL:     (196, 197),
        }.get(level, (33, 39))
        color   = "\033[38;5;{:d}m"         # 256-color to give wider spectrum than just ANSI
        reset   = "\033[0m"

        # Pass any simple messages from internal things, like Django's runserver, without special formatting.
        mp_levels = {
            logging.INFO: u'INF',
            logging.WARNING: u'WRN',
            logging.ERROR: u'ERR',
            logging.DEBUG: u'DBG',
            logging.CRITICAL: u'CRT'
        }
        st_level = mp_levels[level]
        level_prefix = '%s[%s] ' % (color.format(level_colors[0]), st_level)
        return u'{level_prefix}{color_normal}{message}{reset}'.format(
            level_prefix    = level_prefix if self.print_level else '',
            message         = message,
            color_normal    = color.format(level_colors[1]),
            reset           = reset
        )

    def llog(self, logging_level, format_string, *args):
        """
        @param logging_level:
            50 = summary/critical
            40 = error
            30 = warning
            20 = info
            10 = debug
        @return:
        """
        LEVELS = {
            logging.CRITICAL,
            logging.ERROR,
            logging.WARNING,
            logging.INFO,
            logging.DEBUG
        }
        if logging_level not in LEVELS:
            logging_level = logging.DEBUG
        message = format_string % args
        if logging_level >= self.logging_level:
            if self.stdout.isatty():
                message = self.color_format(logging_level, message)
                self.stdout.write(message)
            else:
                self.stdout.write(message)
            self.stdout.write('\n')
            self.add_message(message)

    def add_message(self, message):
        if self.current_line:
            self.output.append((self.current_line + message).strip())
            self.current_line = ''
        elif message is not None:
            self.output.append(message.strip())

    def log(self, format_string, *args):
        message = format_string % args
        self.stdout.write(message)
        self.stdout.write('\n')
        self.add_message(message)

    def color_log(self, fn, format_string, *args):
        message = format_string % args
        if message[-1] == '\n':
            close_line = True
            message = message[:-1]
        else:
            close_line = False
        message = fn(message)
        self.stdout.write(message)
        if close_line:
            self.stdout.write('\n')
            self.add_message(message)
        elif self.current_line:
            self.current_line += message
        else:
            self.current_line = message

    def plain_log(self, format_string, *args):
        self.color_log(lambda x: x, format_string, *args)

    def critical(self, format_string, *args):
        self.llog(logging.CRITICAL, format_string, *args)

    def debug(self, format_string, *args):
        self.llog(logging.DEBUG, format_string, *args)

    def info(self, format_string, *args):
        self.llog(logging.INFO, format_string, *args)

    def warning(self, format_string, *args):
        self.llog(logging.WARNING, format_string, *args)

    def error(self, format_string, *args):
        self.llog(logging.ERROR, format_string, *args)

    def exception(self, format_string, *args):
        p_type, p_exception, p_traceback = sys.exc_info()
        self.llog(logging.ERROR, format_string, *args)
        self.llog(logging.ERROR, u'Exception message: %s', p_exception)
        self.llog(logging.ERROR, u'Exception type   : %s', p_type)
        self.llog(logging.ERROR, u'Traceback\n%s', p_traceback.format_exc())

    def show_last(self):
        try:
            p = CommandOutput.objects.filter(
                command__name=self.command_name
            ).latest('id')
            self.log('%s', p.output)
        except CommandOutput.DoesNotExist:
            self.log('This is the first run of %s', self.command_name)

    def execute(self, *args, **options):
        if options.get('b_show_last', False):
            self.show_last()
            return
        success = False
        try:
            self.start_run()
            super(StatusCommand, self).execute(*args, **options)
            success = True
        except:
            success = False
            raise
        finally:
            self.end_run(success)




import os
import psutil
import resource


__all__ = [ 'get_memory_usage', 'get_peak_memory_usage' ]


def get_memory_usage():
    process = psutil.Process(os.getpid())
    return process.memory_info().rss


def get_peak_memory_usage():
    return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss



#encoding: utf-8

import csv
import cStringIO


class UnicodeCsvWriter(object):
    def __init__(self, f, dialect=csv.excel, **kwds):
        self.queue = cStringIO.StringIO()
        self.writer = csv.writer(self.queue, dialect=dialect, **kwds)
        self.stream = f

    def writerow(self, row):
        """
        writerow(unicode) -> None
        This function takes a Unicode string and encodes it to the output.
        """
        self.writer.writerow([s.encode('utf-8') if isinstance(s, unicode) else s for s in row ])
        data = self.queue.getvalue()
        data = data.decode('utf-8')
        self.stream.write(data)
        self.queue.truncate(0)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)





import codecs
import csv


class Utf8Recoder(object):
    """
    Iterator that reads an encoded stream and reencodes the input to UTF-8
    """
    def __init__(self, f):
        super(Utf8Recoder, self).__init__()
        self.reader = f

    def __iter__(self):
        return self

    def next(self):
        return self.reader.next().encode('utf-8')


class UnicodeCsvReader(object):
    """
    A CSV reader which will iterate over lines in the CSV file "f",
    which is encoded in the given encoding.
    """
    def __init__(self, f, dialect=csv.excel, **kwargs):
        super(UnicodeCsvReader, self).__init__()
        f = Utf8Recoder(f)
        self.reader = csv.reader(f, dialect=dialect, **kwargs)

    def next(self):
        row = self.reader.next()
        return [ unicode(s, 'utf-8') for s in row ]

    def __iter__(self):
        return self


def read_unicode_csv(filename, encoding='utf-8'):
    with codecs.open(filename, 'r', encoding=encoding) as f:
        reader = UnicodeCsvReader(f)
        for x in reader:
            yield x




import inspect
import sys
from importlib import import_module
from django.conf import settings
from django.core.management.base import BaseCommand
from djenga.test import IntegrationTest


class Command(BaseCommand):
    """
    The IntegrationTestCommand is a maangement command
    that will run all classes in any INSTALLED_APP's
    `integration_test` submodule or the specified module
    so long as the class is a subclass of
    `djenga.test.IntegrationTest`.
    """
    help = (
        'run_integration_tests is a management command '
        "that will run all classes in any INSTALLED_APP's "
        '`integration_test` submodule or the specified module '
        'so long as the class is a subclass of '
        '`djenga.test.IntegrationTest`.')

    def __init__(self, stdout=None, stderr=None, no_color=False):
        super(Command, self).__init__(stdout, stderr, no_color)
        self.modules = None
        """@type: list[str] list of dot delimited modules in which to look for tests"""
        self.tests = None
        """@type: list[str] list of class names of specific tests wanted to be run"""

    def create_parser(self, prog_name, subcommand):
        parser = super(Command, self).create_parser(prog_name, subcommand)
        parser.add_argument(
            '--module', '-m',
            action='append',
            dest='modules',
            help='specifies a dotted module path in which '
                 'one or more `IntegrationTest` subclasses '
                 'can be found and run')
        parser.add_argument(
            '--test', '-t',
            action='append',
            dest='tests',
            help='specifies the name of '
                 'one or more `IntegrationTest` subclasses '
                 'to be run')
        return parser

    def check_module(self, module, display_warnings=True):
        try:
            import_module(module)
            return True
        except ImportError:
            if display_warnings:
                self.stdout.write('WARNING: could not import module [%s]' % module)
            return False

    def set_modules(self, modules):
        display_warnings = False
        if modules:
            display_warnings = True
        else:
            modules = [
                '%s.integration_tests' % x
                for x in settings.INSTALLED_APPS ]
        self.modules = [
            x for x in modules
            if self.check_module(x, display_warnings) ]

    def set_tests(self, tests):
        self.tests = []
        for module in self.modules:
            module = import_module(module)
            for name, klass in inspect.getmembers(module):
                if name.startswith('__') or not inspect.isclass(klass): continue
                x = klass()
                if isinstance(x, IntegrationTest) and (
                    not tests or
                    klass.__name__ in tests
                ):
                    self.tests.append(x)

    def run_tests(self):
        all_passed = True
        for x in self.tests:
            status = x.run_test()
            all_passed = status and all_passed
        return all_passed

    def handle(self, *args, **options):
        self.set_modules(options.get('modules'))
        self.set_tests(options.get('tests'))
        success = self.run_tests()
        if not success:
            sys.exit(1)




from django.db import models


__all__ = [
    'ManagementCommand',
    'CommandOutput',
]


class ManagementCommand(models.Model):
    name = models.CharField(
        max_length=128,
        null=False,
        blank=False,
        db_index=True,
        help_text='The full path of the command',
    )
    status = models.CharField(
        max_length=32,
        null=False,
        blank=False,
        help_text='The current status of the command',
    )
    last_run = models.DateTimeField(
        null=True,
        help_text='The last time this command was run',
    )
    last_success = models.DateTimeField(
        null=True,
        help_text='The last time this command was run successfully',
    )
    created_at = models.DateTimeField(null=False, auto_now_add=True)
    updated_at = models.DateTimeField(null=False, auto_now=True)


class CommandOutput(models.Model):
    command = models.ForeignKey(ManagementCommand)
    output = models.TextField(null=True)
    created_at = models.DateTimeField(null=False, auto_now_add=True)
    updated_at = models.DateTimeField(null=False, auto_now=True)



from __future__ import print_function
import sys


def flush_print(st, *args, **kwargs):
    end = kwargs.pop('end', '\n')
    print(st % args, end=end)
    if sys.stdout.isatty():
        sys.stdout.flush()



# encoding: utf-8
# pylint: disable=pointless-string-statement

from collections import Iterable
from datetime import datetime
from datetime import date
from decimal import Decimal
import json
from dateutil import parser as date_parser
from django.db import models


__all__ = [
    'JsonMixin',
    'to_json',
    'from_json',
]


def make_json_serializable(value, fn=lambda x: x):
    if value is None:
        return None
    if isinstance(value, datetime):
        return value.isoformat()
    if isinstance(value, date):
        return value.isoformat()
    if isinstance(value, Decimal):
        return float(value)
    return fn(value)


class JsonEncoder(json.JSONEncoder):
    def default(self, value):
        return make_json_serializable(value, super(JsonEncoder, self).default)


def pairs_hook(pairs):
    data = {}
    for key, value in pairs:
        if not isinstance(value, basestring):
            data[key] = value
        else:
            try:
                data[key] = date_parser.parse(value)
            except ValueError:
                data[key] = value
    return data


def to_json(p, additional_fields=None):
    if not isinstance(p, models.Model):
        raise ValueError('The record provided is not a models.Model instance')
    fn = getattr(p._meta, 'get_fields', None)
    data = {}
    if fn:
        for x in fn():
            field_name = x.name + '_id' if isinstance(x, models.ForeignKey) else x.name
            data[field_name] = make_json_serializable(getattr(p, field_name, None))
    else:
        # Handle pre-django 1.9
        for x in p._meta.get_all_field_names():
            try:
                field = p._meta.get_field(x)
                field_name = x + '_id' if isinstance(field, models.ForeignKey) else x
                data[field_name] = make_json_serializable(getattr(p, field_name, None))
            except models.FieldDoesNotExist:
                data[x] = getattr(p, x, None)
    if additional_fields and isinstance(additional_fields, Iterable):
        for x in additional_fields:
            data[x] = getattr(p, x, None)
    return data


def from_dict(p, data):
    for key, value in data.iteritems():
        setattr(p, key, value)


def from_json(p, data):
    data = json.loads(data, object_pairs_hook=pairs_hook)
    for key, value in data.iteritems():
        setattr(p, key, value)


class JsonMixin(object):
    def to_json(self, additional_fields=None):
        return to_json(self, additional_fields)

    def from_json(self, data):
        pass


from setuptools import setup, find_packages
import sys, os

version = '0.3.6'

setup(name='djenga',
      version=version,
      description="Useful building blocks for Django.",
      long_description="""\
Useful building blocks for Django.""",
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Developers',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Framework :: Django',
          'Framework :: Django :: 1.11',
          'Framework :: Django :: 1.10',
          'Framework :: Django :: 1.9',
          'License :: OSI Approved :: BSD License',
      ], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='django python',
      author='Preetam Shingavi',
      author_email='p.shingavi@yahoo.com',
      url='https://github.com/2ps/djenga',
      license='BSD',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=True,
      install_requires=[
          'Django>=1.9',
          'cssutils==1.0',
          'beautifulsoup4==4.3.2',
          'pycrypto>=2.6.1',
          'pytz',
          'python-dateutil',
          'pyyaml',
      ] + ([ 'psutil' ] if 'cygwin' in sys.platform else []),
      entry_points="""
      # -*- Entry points: -*-
      """,
      )




from nested_transactions import commit_on_success
from nested_transactions import start_transaction
from nested_transactions import commit_transaction
from nested_transactions import rollback_transaction
from timestampfield import TimestampField






from django.db import transaction
from django.db import DEFAULT_DB_ALIAS
from functools import wraps
import logging
import threading


logger = logging.getLogger(__name__)
__all__ = [
    u'start_transaction',
    u'end_transaction',
    u'rollback_transaction',
    u'commit_transaction',
]
# transaction_data is a thread local data structure that will
# make sure that we are operating on the proper transaction
# stack when saving transaction depth and SIDs.  Note, that
# the limitation of this design is that you cannot use multiple
# threads to operate on a single transaction.
transaction_data = threading.local()


class HandleTransactons(object):
    def __init__(self, using):
        self.using = using
        self.function_name = None

    def __enter__(self):
        start_transaction(self.using)

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            if exc_val is not None:
                logger.exception(u'Got an exception trying to execute %s; rolling back transaction: %s',
                                 self.function_name,
                                 exc_val)
                rollback_transaction(self.using)
            else:
                commit_transaction(self.using)
        finally:
            end_transaction(self.using)

    def __call__(self, func):
        @wraps(func)
        def inner(*args, ** kwargs):
            self.function_name = func.__name__
            with self:
                return func(*args, **kwargs)
        return inner


def start_transaction(using=None):
    """
    Starts transaction management in a way that is safe for nested transactions
    on platforms such as MySQL.  If the transaction depth is greater than 1, the
    start_transaction() function will use savepoints.  Otherwise, the transaction
    is treated normally.  The start_transaction() pushes the savepoint SID to a
    thread local data structure to keep track of current depth and active savepoints.
    Anytime end_transaction() is called, the most recent savepoint, if any, will
    be popped off the stack.
    @param using:  The database alias string.  The using argument defaults to None, and
    if None is provided, the function operates as if using were set to
    django.db.transaction.DEFAULT_DB_ALIAS
    @return: Nothing
    """
    if using is None:
        using = DEFAULT_DB_ALIAS
    if not hasattr(transaction_data, u'transaction_depth'):
        transaction_data.transaction_depth = dict()
    if not hasattr(transaction_data, u'sids'):
        transaction_data.sids = dict()

    transaction_depth = transaction_data.transaction_depth
    current_depth = transaction_depth.get(using, 0)
    current_depth += 1
    transaction_depth[using] = current_depth
    #logger.debug(u'Current transaction depth: %d', current_depth)
    if current_depth == 1:
        transaction.enter_transaction_management(managed=True, using=using)
        transaction.managed(flag=True, using=using)
    else:
        sid = transaction.savepoint(using=using)
        if not using in transaction_data.sids:
            transaction_data.sids[using] = list()
        transaction_data.sids[using].append(sid)


def end_transaction(using=None):
    """
    Ends transaction management for a given frame, with proper accounting
    for nesting.  This function will check the current transaction depth for
    the given using parameter.  The function will end transaction management
    if the transaction depth is 0.
    @param using: A string object for the database name.  If None is provided as
    the argument, then the using parameter defaults to transaction.DEFAULT_DB_ALIAS.
    @type using: str
    @return: None
    """
    if using is None:
        using = DEFAULT_DB_ALIAS
    if not hasattr(transaction_data, u'transaction_depth'):
        raise BaseException(u'Call to end_transaction() made without an active transaction')
    transaction_depth = transaction_data.transaction_depth
    current_depth = transaction_depth.get(using, 0)
    if current_depth == 0:
        raise BaseException(u'Call to end_transaction() made without an active transaction')
    current_depth -= 1
    if current_depth == 0:
        transaction.leave_transaction_management(using=using)
        del transaction_data.transaction_depth[using]
        if using in transaction_data.sids:
            del transaction_data.sids[using]
    else:
        transaction_depth[using] = current_depth


def rollback_transaction(using=None):
    if using is None:
        using = DEFAULT_DB_ALIAS
    transaction_depth = transaction_data.transaction_depth
    current_depth = transaction_depth.get(using, 0)
    if current_depth > 1:
        sid = transaction_data.sids[using].pop()
        transaction.savepoint_rollback(sid, using=using)
    else:
        transaction.rollback(using=using)


def commit_transaction(using=None):
    if using is None:
        using = DEFAULT_DB_ALIAS
    transaction_depth = transaction_data.transaction_depth
    current_depth = transaction_depth.get(using, 0)
    if current_depth > 1:
        sid = transaction_data.sids[using].pop()
        transaction.savepoint_commit(sid, using=using)
    else:
        transaction.commit(using=using)


def commit_on_success(using=None):
    """
    Default decorator that mimics the style of the django.db.transaction
    decorator of the same name.  Because the HandleTransactions class takes
    optional arguments, this decorator has to perform the "callable" magic
    so that the decorator operates as expected.
    @param using: The database alias associated with the function.  The function
    will default to using transaction.DEFAULT_DB_ALIAS if no value is provided.
    @return: the decorated function
    """
    if using is None:
        using = DEFAULT_DB_ALIAS
    if callable(using):
        return HandleTransactons(DEFAULT_DB_ALIAS)(using)
    return HandleTransactons(using=using)




from commandlogginglevels import VerbosityLevels
from configurationmixin import ConfigurationMixin
from csvmixin import CsvMixin
from inputcsvmixin import InputCsvMixin
from statuscommand import StatusCommand
from run_integration_tests import Command as IntegrationTestCommand

__all__ = [
    'VerbosityLevels',
    'ConfigurationMixin',
    'CsvMixin',
    'InputCsvMixin',
    'StatusCommand',
    'IntegrationTestCommand',
]




import importlib


def get_function(name):
    """
    Loads a function by its fully qualified name.  Examples:
    `get_function('datetime.datetime')`
    `get_function('djenga.reflection.get_function')`

    :param name: the fully-qualified function name
    :type name: str | unicode
    :return: the function requested
    :rtype: function | class
    """
    if name in locals():
        return locals()[name]
    if name in globals():
        return globals()[name]
    st_module, st_function = name.rsplit('.', 1)
    module = importlib.import_module(st_module)
    return getattr(module, st_function, None)
